(define @@documentation '())
(define-macro (@doc symbol . arguments)
  (set! @@documentation
    (cons (cons symbol arguments)
	  @@documentation))
  #t)

(define (@help-symbol symbol)
  (define (display-section section) 
    (display section)
    (display ": "))
  (define (displayn text)
    (display text)
    (newline))

  (define (get-parameter-list params)
    (map (lambda (p)
	   (cond ((list? p) (cadr p))
		 ((equal? p '@etc) "...")
		 (else p)))
	 params))
  (define (get-parameter-descs params)
    (apply append
	   (map (lambda (p)
		  (if (list? p) (list p) '()))
		params)))


  (let ((doc (assoc symbol @@documentation)))
    (if doc
      (let ((desc (assoc 'desc (cdr doc)))
	    (params (assoc 'params (cdr doc)))
	    (returns (assoc 'returns (cdr doc)))
	    (example (assoc 'example (cdr doc)))
	    (notes (assoc 'notes (cdr doc))))
	(let ((parameter-list (get-parameter-list (cdr params)))
	      (parameter-desc (get-parameter-descs (cdr params))))
	(display "Documentation for: (") 
	(display symbol) 
	(map (lambda (p)
	       (display " ") (display p))
	     parameter-list)
	(display ")")
	(newline)(newline)

	(if desc (begin (map displayn (cdr desc)) (newline)))
	(if (not (null? parameter-desc))
	  (begin
	    (display "Parameter(s):") (newline)
	    (map (lambda (p)
		   (display "\t") (display (car p)) (display " - ") (display (cadr p))
		   (newline))
		 parameter-desc)))
	
	(if returns (begin (display-section "Returns") (map displayn (cdr returns))))
	(if example (begin (display-section "Example(s)") (newline) (map displayn (cdr example)) (newline)))
	(if notes (begin (display-section "Notes") (map displayn (cdr notes))))))
      (begin
	(display "No documentation for ") (display symbol) (display " found.")
	(newline))))
  (void))
	    
(define-macro (@help symbol)
  `(@help-symbol ',symbol))	   

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;; numeric operations
;;
(@doc + (params z1 @etc)
      (desc "This procedure returns the sum of its arguments.")
      (example "(+ 3 4 5)  ===>  12"
	       "(+ 3 4)    ===>  7"
	       "(+ 3)      ===>  3"
	       "(+)        ===>  0")
      (notes R5RS))

(@doc * (params z1 @etc)
      (desc "This procedure returns the product of its arguments.")
      (example "(* 3 4 5)  ===>  60"
	       "(* 3 4)    ===>  12"
	       "(* 3)      ===>  3"
	       "(*)        ===>  1")
      (notes R5RS))

(@doc - (params z1 @etc)
      (desc "With two or more arguments, this procedure returns the difference of its arguments, associating to the left. With one argument, however, it returns the additive inverse of its argument.")
      (example "(- 3 4)    ===>  -1"
	       "(- 3 4 5)  ===>  -6"
	       "(- 3)      ===>  -3")
      (notes R5RS))

(@doc / (params z1 @etc)
      (desc "With two or more arguments, this procedure returns the quotient of its arguments, associating to the left. With one argument, however, it returns the multiplicative inverse of its argument.")
      (example "(/ 3 4)    ===>  0.75"
	       "(/ 3 4 5)  ===>  0.15"
	       "(/ 3)      ===>  0.33333")
      (notes R5RS))

;;
;; list operations
;;
(@doc append (params list @etc)
      (desc "Returns a list consisting of the elements of the first list followed by the elements of the other lists.")
      (example "(append '(x) '(y))        ===>  (x y)"
	       "(append '(a) '(b c d))    ===>  (a b c d)"
	       "(append '(a (b)) '((c)))  ===>  (a (b) (c))"))

(@doc reverse (params list)
      (desc "Returns a newly allocated list consisting of the elements of list in reverse order.")
      (example "(reverse '(a b c))              ===>  (c b a)"
      	       "(reverse '(a (b c) d (e (f))))  ===>  ((e (f)) d (b c) a)")
      (notes R5RS))

(@doc list (params obj @etc)
      (desc "Returns a newly allocated list of its arguments.")
      (example "(list 'a (+ 3 4) 'c)                    ===>  (a 7 c)"
	       "(list)                                  ===>  ()")
      (notes R5RS))

(@doc null? (params obj)
      (desc "Returns #t if obj is the empty list, otherwise returns #f.")
      (notes R5RS))

(@doc car (params pair)
      (desc "Returns the contents of the car field of pair. Note that it is an error to take the car of the empty list.")
      (example 	"(car '(a b c))                  ===>  a"
      		"(car '((a) b c d))              ===>  (a)"
      		"(car '(1 . 2))                  ===>  1"
      		"(car '())                       ===>  error")
      (notes R5RS))

(@doc cdr (params pair)
      (desc "Returns the contents of the cdr field of pair. Note that it is an error to take the cdr of the empty list.")
      (example 	"(cdr '((a) b c d))              ===>  (b c d)"
		"(cdr '(1 . 2))                  ===>  2"
		"(cdr '())                       ===>  error")
      (notes R5RS))

(@doc cons (params obj1 obj2)
      (desc "Returns a newly allocated pair whose car is obj1 and whose cdr is obj2. The pair is guaranteed to be different (in the sense of eqv?) from every existing object.")
      (example  "(cons 'a '())                   ===>  (a)"
		"(cons '(a) '(b c d))            ===>  ((a) b c d)"
		"(cons 'a 3)                     ===>  (a . 3)"
		"(cons '(a b) 'c)                ===>  ((a b) . c)")
      (notes R5RS))

(@doc member (params obj list)
      (desc "This procedure returns the first sublist of list whose car is obj, where the sublists of list are the non-empty lists returned by (list-tail list k) for k less than the length of list. If obj does not occur in list, then #f (not the empty list) is returned. Member uses equal? to compare obj with the elements of list.")
      (example "(member (list 'a) '(b (a) c))                     ===>  ((a) c)")
      (notes R5RS))

(@doc length (params list)
      (desc "Returns the length of list.")
      (example "(length '(a b c))                       ===>  3"
	       "(length '(a (b) (c d e)))               ===>  3"
	       "(length '())                            ===>  0")
      (notes R5RS))

(@doc set-car! (params pair obj)
      (desc "Stores obj in the car field of pair. The value returned by set-car! is unspecified.")
      (example "(define (f) (list 'not-a-constant-list))"
	       "(define (g) '(constant-list))"
	       "(set-car! (f) 3)                     ===>  unspecified value"
	       "(set-car! (g) 3)                     ===>  unspecified behavior (in most cases works as expected)")
      (notes side-effect STM))

(@doc set-cdr! (params pair obj)
      (desc "Stores obj in the cdr field of pair. The value returned by set-cdr! is unspecified.")
      (example "(define (f) (list 'not-a-constant-list))"
	       "(define (g) '(constant-list))"
	       "(set-cdr! (f) 3)                     ===>  unspecified value"
	       "(set-cdr! (g) 3)                     ===>  unspecified behavior (in most cases works as expected)")
      (notes side-effect STM))


;; demo
;(@doc quotient
;      (params 
;	(n1 "numerator")
;	(n2 "denominator"))
;      (desc     "This procedure implements number-theoretic (integer) division. n2 should be non-zero.")
;      (returns 	"This procedure returns integer.")
;      (example 	"(quotient n1 n2) ===> n1/n2")
;      (notes 	R5RS))
