;;;;;
;;;;;  Matrix Operations
;;;;;

;; get value -- unsafe code (vector-ref can overflow)
(define (matrix-ref matarg i j)
  (vector-ref (caddr matarg) (+ (* i (cadr matarg)) j)))

;;;;;
;;;;;  Naive FCA Algorithm
;;;;;

;; test property of all list members
(define (for-all f l)
  (or (null? l)
      (and (f (car l))
	   (for-all f (cdr l)))))

;; fast my-filter
(define (my-filter f l)
  (cond ((null? l) '())
	((f (car l)) (cons (car l) (my-filter f (cdr l))))
	(else (my-filter f (cdr l)))))

;; adjust truth-degree representation
(define (adjust elem t)
  (cond ((= t 0) '())
	((= t 1) (list elem))
	(else (list (cons elem t)))))

;; return truth degree
(define (degree x) (if (pair? x) (cdr x) 1))

;; return element
(define (elem x) (if (pair? x) (car x) x))

;; meet on [0,1]
(define (meet vals)
  (if (null? vals)
      1
      (apply min vals)))

;; general arrow-operator
(define (arrow A res I y-set ref)
  (apply append
	 (map (lambda (y)
		(let ((result 
		       (meet (map res 
				  (map (lambda (x)
					 (cons (degree x) 
					       (ref I (elem x) y))) 
				       A)))))
		  (adjust y result)))
	      y-set)))

;; general truth-stresser
(define (ts set truth-stresser)
  (apply append
	 (map (lambda (x) 
		(adjust (elem x) 
			(truth-stresser (elem x) (degree x)))) 
	      set)))

;; arrow-up operator
(define (up A res I y-set)
  (arrow A res I y-set (lambda (I x y) (matrix-ref I (elem x) y))))

;; arrow-down operator
(define (dn B res I x-set)
  (arrow B res I x-set (lambda (I x y) (matrix-ref I y (elem x)))))

; (up dn up ts ts up dn)

;; create a range
(define (range n)
  (define (iter n aux)
    (if (= n 0)
	(cons 0 aux)
	(iter (- n 1) (cons n aux))))
  (iter n '()))

;; galois eval
(define (galois-eval expr set res truth-stresser cont)
  (let ((obj (range (- (car cont) 1)))
	(atr (range (- (cadr cont) 1))))
    (define (iter expr aux)
      (cond ((null? expr) aux)
	    ((equal? (car expr) 'up) (iter (cdr expr) (up aux res cont atr)))
	    ((equal? (car expr) 'dn) (iter (cdr expr) (dn aux res cont obj)))
	    ((equal? (car expr) 'ts) (iter (cdr expr) (ts aux truth-stresser)))
	    (else (list "Error: unknown operator or something."))))
    (iter expr set)))


;;;;;
;;;;;  General fixpoint utilities
;;;;;

;; belongs
(define (belongs A x)
  (apply max
	 (cons 0
	       (map (lambda (y)
		      (if (equal? (elem x) (elem y))
			  (degree y)
			  0))
		    A))))

;; lexical strict order
(define (leq-lex a b)
  (or (< (car a) (car b))
      (and (= (car a) (car b))
	   (<= (cdr b) (cdr a)))))

;; <_i strict-ordering
(define (lt-i A B i)
  (define (iter n)
    (if (= n i)
	(< (belongs A n) (belongs B n))
	(and (= (belongs A n) (belongs B n))
	     (iter (+ n 1)))))
  (iter 0))

;; (B \cap \{0,\dots,i-1}) \cup \{j/i\}
(define (lex-add B i j)
  (define (iter n)
    (if (= n i)
	(adjust i j)
	(append (adjust n (belongs B n))
		(iter (+ n 1)))))
  (iter 0))

;; l-set subsethood predicate
(define (lset-subset? A B)
  (for-all (lambda (x)
	     (<= (degree x) (belongs B (elem x))))
	   A))

;; l-set equality predicate
(define (lset-equal? A B)
    (and (lset-subset? A B)
	 (lset-subset? B A)))

;; limit operator
(define (limit clos)
  (define (rec A)
    (let ((result (clos A)))
      (if (lset-equal? A result)
	  A
	  (rec (clos A)))))
  rec)

;; get the least fixpoint
(define (least-fixpoint B clos max-i j-set)
  (define (lex-max)
    (cons max-i j-set))
  (define (lex-pred ij)
    (cond ((null? (cdr ij)) (error "This should never happen: j-val = ()"))
	  ((null? (cddr ij)) (cons (- (car ij) 1) j-set))
	  (else (cons (car ij) (cddr ij)))))
  (define (iter ij)
    (if (null? ij)
	(error "This should not happen: ij = 0")
	(let ((result (clos (lex-add B (car ij) (cadr ij)))))
	  (if (lt-i B result (car ij))
	      result
	      (iter (lex-pred ij))))))
  (iter (lex-max)))

;; get all fixed points
(define (all-fixpoints start clos elems truths)
  (define all-res (range elems))
  (define (iter last accum)
    (if (equal? last all-res)
	(reverse accum)
	(let ((next (least-fixpoint last clos elems truths)))
	  (iter next (cons next accum)))))
  (append (if (equal? start (clos start))
	      (list start)
	      '())
	  (iter start '())))

;; lukasiwicz residuum
(define (luk-res vals)
  (min 1 (+ (- 1 (car vals)) (cdr vals))))

;; minimum residuum
(define (min-res vals)
  (if (<= (car vals) (cdr vals)) 1 (cdr vals)))

;; Baaz's truth-stresser
(define (baaz-ts elem degree)
  (if (= degree 1) 1 0))

;; small, large, far, near
(define context
  (list 72 4
	#(
	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto


	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto


	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto

	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto

	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto


	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto


	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto

	  1     0    0     1      ; 0: Merkur
	  0.75  0    0     1      ; 1: Venuse
	  0.75  0    0     0.75   ; 2: Zeme
	  1     0    0.5   0.75   ; 3: Mars
	  0     1    0.75  0.5    ; 4: Jupiter
	  0     1    0.75  0.5    ; 5: Saturn
	  0.25  0.5  1     0.25   ; 6: Uran
	  0.25  0.5  1     0      ; 7: Neptun
	  1     0    1     0	  ; 8: Pluto



	  )))

(display (length (all-fixpoints '()
	       (lambda (B)
		 (galois-eval '(dn ts up) B luk-res baaz-ts context))
	       3 '(0 0.25 0.5 0.75 1))))
